/*
  TFM-CUDA
  Copyright (C) 2008-2010  Mathieu Giraud & Jean-Stéphane Varré
  http://bioinfo.lifl.fr/TFM/TFM-CUDA/ 

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


void init_cuda(int device)
{
  int deviceCount, dev;
  
  cudaGetDeviceCount(&deviceCount);
  
  if (deviceCount == 0)
    {
      perror("No CUDA device");
      exit(9);
    }
	
  printf("\n");

  for (dev = 0; dev < deviceCount; ++dev)
    {
      cudaDeviceProp deviceProp;
      cudaGetDeviceProperties(&deviceProp, dev);
      
      printf("==[%d]==  %s\n", dev, deviceProp.name );

      if (conf_verbose)
      {
      printf("cc       %d.%d\n", deviceProp.major, deviceProp.minor);
      printf("Freq     %d MHz\n", deviceProp.clockRate / 1000);
      printf("Memory   global: %zd bytes (%ld MB)\n", deviceProp.totalGlobalMem, (long) deviceProp.totalGlobalMem / (1 << 20)); 
      printf("         const: %zd bytes\n", deviceProp.totalConstMem); 
      printf("         shared : %zd bytes / block\n", deviceProp.sharedMemPerBlock ); 
      printf("         regs: %d / block\n", deviceProp.regsPerBlock ); 
      printf("warpSize %d\n", deviceProp.warpSize );
      printf("Limits   threads: <%d, %d, %d>\n", deviceProp.maxThreadsDim[0], deviceProp.maxThreadsDim[1], deviceProp.maxThreadsDim[2]); 
      printf("         threads per block: %d\n", deviceProp.maxThreadsPerBlock); 
      printf("         grid: <%d, %d, %d>\n", deviceProp.maxGridSize[0], deviceProp.maxGridSize[1], deviceProp.maxGridSize[2]); 
      printf("         memPitch: %zd\n", deviceProp.memPitch );
      }
      printf("\n");      
    }

  printf ("== Initializing device ==[%d]==\n", device);

  if (cudaSetDevice(device))
    {
      perror("CUDA Device init failed.");
      exit(4);
    }

#ifdef  __DEVICE_EMULATION__
  printf("==[x]== Using Device Emulation (CPU)\n\n");
#endif

  time_stop();
}

