/*
  TFM-CUDA
  Copyright (C) 2008-2010  Mathieu Giraud & Jean-Stéphane Varré
  http://bioinfo.lifl.fr/TFM/TFM-CUDA/ 

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


__device__ void copy_float_device_shared(float *shared, float *device, int length)
{
  {
    int i;
#ifdef PARALLEL_COPY
    for (i=threadIdx.x; i<length; i+= NB_THREADS_PER_BLOCK)
#else
    for (i=0; i<length; i++)
#endif
	shared[i] = device[i] ;
  }


#ifdef PARALLEL_COPY
  __syncthreads();
#endif
}

__device__ void copy_int_device_shared(int *shared, int *device, int length)
{
  {
    int i;
#ifdef PARALLEL_COPY
    for (i=threadIdx.x; i<length; i+= NB_THREADS_PER_BLOCK)
#else
    for (i=0; i<length; i++)
#endif
	shared[i] = device[i] ;
  }

#ifdef PARALLEL_COPY
  __syncthreads();
#endif
}


__device__ void copy_uchar_device_shared(unsigned char *shared, unsigned char *device, int length)
{
  {
    int i;
#ifdef PARALLEL_COPY
    for (i=threadIdx.x; i<length; i+= NB_THREADS_PER_BLOCK)
#else
    for (i=0; i<length; i++)
#endif
	shared[i] = device[i] ;
  }

#ifdef PARALLEL_COPY
  __syncthreads();
#endif
}


__device__ void copy_seq_device_shared(int *shared, unsigned char *device, int length)
{
  
#ifdef PAD
   int *dev = (int*) device ;
#endif

    int i;
    __syncthreads();

#ifdef PARALLEL_COPY
    for (i=threadIdx.x; i<length; i+= 
         #ifdef PAD
		 4*
         #endif
		 NB_THREADS_PER_BLOCK)
#else
    for (i=0; i<length; i++)
#endif

    {
#ifdef PAD     // not working now (10% time)
	int buf = dev[i] ;
	unsigned char *b = (unsigned char *) &buf ;
       	shared[i+0] = (int) b[0] ;
	shared[i+1] = (int) b[1] ;
	shared[i+2] = (int) b[2] ;
	shared[i+3] = (int) b[3] ;
#else
	shared[i] = device[i] ;
#endif	
    }


#ifdef PARALLEL_COPY
  __syncthreads();
#endif
}
