/*
  TFM-CUDA
  Copyright (C) 2008-2010  Mathieu Giraud & Jean-Stéphane Varré
  http://bioinfo.lifl.fr/TFM/TFM-CUDA/ 

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <math.h>


__device__ void printf_word(int *word, int length)
{
  int t ;
  for (t=0; t<length; t++)
    PRINTF("%c", "ACGTacgt"[word[t]]);
  
  PRINTF(" ");
}

void printf_matrix(float *matrix, int length)
{
  int i ;

  if (conf_verbose)
  for (i=0; i<ALPH_SIZE*length; i++)
    {
      printf(" %8.3f ", matrix[i]);
      if (!((i+1)%ALPH_SIZE))
	printf("\n");

      if (i==8*ALPH_SIZE+1)
      { 
	      printf ("    ...\n");
	      break ;
      }
    }

  printf("== <-- length: %d\n\n", length);
}

void printf_seq(unsigned char *seq, int length)
{
  int i ;
  for (i=0; i<length; i++)
    {
      printf("%d", (int) seq[i]);
    }

  // printf("\n== <-- length: %d\n\n", length);
}

float *random_matrix(int length)
{
  printf(" <-- random matrix\n");

  float *matrix = (float *) malloc(ALPH_SIZE * length * sizeof(float));

  int i ;
  for (i=0; i<ALPH_SIZE*length; i++)
    matrix[i] = (1 << 2) - ((float) random()) / (1 << 28) ;

  printf_matrix(matrix, length);

  return matrix ;
}


unsigned char *random_seq(int length)
{
  printf(" <-- random seq: %.3e\n", (float) length);

  unsigned char *seq = (unsigned char*) malloc(sizeof(unsigned char) * length);

  int i ;
  for (i=0; i<length; i++)
    seq[i] = (random() % ALPH_SIZE) ;

  // printf_seq(seq, length);

  return seq ;
}

unsigned char *seq_from_file(const char *fname, int *length)
{
  printf(" <-- %s\n", fname);

	int code[256] ;
	int i ;
	for (i=0; i<255; i++)
		code[i] = 9 ;

	code['A'] = code['a'] = 0 ;
	code['C'] = code['c'] = 1 ;
	code['G'] = code['g'] = 2 ;
	code['T'] = code['t'] = 3 ;


	FILE* f = fopen(fname, "r");
	if (f == 0)
	{
		perror(fname);
		exit(79);
	}

  unsigned char *seq = (unsigned char*) malloc(sizeof(unsigned char) * 500000000);
	char c ;
	*length = 0 ;

  while ((c = fgetc(f)) != EOF)
	{
		int n = code[c];
		if (n == 9)
			continue ;

		seq[(*length)++] = n;
	}

	printf(" <-- length: %d\n", *length);

  return seq ;
}


float matrix_buf[ALPH_SIZE*MATRIX_MAX_LENGTH];

float *matrix_from_file(float* matrix_force, const char *fname, int *length)
{
  printf("== <-- %s\n", fname);
  FILE* f = fopen(fname, "r");

  if (f == 0)
    {
      perror (fname);
      exit(5);
    }


  // Read in a buffer
  int n = 0 ;
  int stop = 0 ;

  while (stop != 1)
    {
      int a ;
      for (a=0; a<ALPH_SIZE; a++)
	{
	  if (fscanf(f, "%f", matrix_buf+n) <= 0)
	    { 
	      if (a+n+stop == 0)
		{
		  stop = 2 ;
		  if (fscanf(f, "> %s\n", (char *) matrix_buf) <=0 )
		    break ;

		  printf (">>> %s\n", (char *) matrix_buf);
		  a-- ;
		  continue ;
		}
	     
	      
	      stop = 1 ;
	      break ;
	    }
	  n++ ;
	}
    }

  *length = n/4 ;

  // Prepare new matrix
  float *matrix = matrix_force ? matrix_force : (float *) malloc(ALPH_SIZE * (*length) * sizeof(float));

  int a ;
  for (n=0; n<*length; n++)
    for (a=0; a<ALPH_SIZE; a++)
      matrix[n*ALPH_SIZE+a] = matrix_buf[a*(*length)+n] ;

  printf_matrix(matrix, *length);

  return matrix; 
}


void display_results(int *results_occ, int *results_mat, int *results_pos, float *results_score, 
		     int length, int nb_matrices, int display_hits)
{
  long  long  nbocc = 0 ;
  long long nbocc_mat[MAX_TOTAL_LENGTH] ;
  int i;
  int results_max = 0 ;
  float sum_score = 0.0;
  int bad_threads = 0 ;

  if (display_hits)
    {
      printf("== HITS ==\n");
      printf("| Matrix nb. ");
      printf("|     Position ");
      printf("|    Score |\n" );
    }

  for (i=0; i<nb_matrices; i++)
    nbocc_mat[i] = 0 ;
 
  for (i=0; i<NB_THREADS_SCAN; i++)
    {
      // printf(" %2d", results_occ[i]);
      nbocc += results_occ[i] ;

      if (results_occ[i] > RES_PER_THREAD)
	{
	  bad_threads++ ;
	  // if (bad_threads < 5)
	  //   printf("==[!]==  Warning: thread %d has %d results, some results will be lost \n", i, results_occ[i]);
	}

      if (results_occ[i] > results_max)
        results_max = results_occ[i];

      int j ;
      for (j=0; j<results_occ[i]; j++)
	{	      
	  nbocc_mat[results_mat[i*RES_PER_THREAD + j]]++ ;
	  
	  if (display_hits)
	    {
	      printf("|");
	      printf("  %9d |", results_mat[i*RES_PER_THREAD + j]);
	      printf(" %12d |", results_pos[i*RES_PER_THREAD + j]);
#ifdef SCORES_OUT
	      printf(" % 8.2f |\n", results_score[i*RES_PER_THREAD + j]);
	      sum_score += results_score[i*RES_PER_THREAD + j];
#else
	      if (j)
		printf("\n");
#endif
	    }
	}
    }

  if (display_hits)
    {
      printf("== END OF HITS ==\n");
    }
  
  for (i=0; i<nb_matrices; i++)
    printf("  Matrix %2d :   %7lld occurences, prob %.6e \n", 
	   i, nbocc_mat[i], (float) nbocc_mat[i] / (float) length);

  if (bad_threads)
    printf("==[!]==  Warning: .... %d threads with too many results, some results are lost.\n", bad_threads);

  if (bad_threads)
    printf("==[!]==  Please use a higher threshold or compile with an higher RES_PER_THREAD.\n");
    
  
  printf("== >>> %lld occurences %f", nbocc, sum_score);
  printf(" (best thread: %d) ", results_max);

#ifdef OLD_PVALUE
  printf(">>> p-value: %.6e\n", (float) nbocc / pow((float) ALPH_SIZE, (float) length)); 
#endif

  printf(">>> prob: %.6e\n", (float) nbocc / (float) length);
}



#define ABS(a)  ((a>0)?(a):-(a))
#define MAX(a,b) ((a)>(b)?(a):(b))
#define MIN(a,b) ((a)>(b)?(b):(a))

__device__ __host__ float max_score_column (float *column) {
	return MAX(MAX(*column, *(column+1)),MAX(*(column+2),*(column+3)));
}

__device__ __host__ float min_score_column (float *column) {
	return MIN(MIN(*column, *(column+1)),MIN(*(column+2),*(column+3)));
}

__device__ __host__ float max_score (float *matrix, int length) {
	float sum = 0;      
	for (int i = 0; i < length; i++) {
		sum += max_score_column (matrix + i * ALPH_SIZE);
	}
	return sum + 1 ;
}

__device__ __host__ float min_score (float *matrix, int length) {
	float sum = 0;      
	for (int i = 0; i < length; i++) {
		sum += min_score_column (matrix + i * ALPH_SIZE);
	}
	return sum;
}
